/*********************************************************************
 *** hatimerun.c
 *** Copyright (c) 2001,2003 by Markus Winand <mws@fatalmind.com>
 *** $Id: hatimerun.c,v 1.15 2003/01/05 14:10:36 mws Exp $
 ********************************************************************/

#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include "hatimerun.h"
#include "signumber.h"
#include "error.h"

extern char* optarg;
extern int optind;

int FAIL_EXITCODE = _FAIL_EXITCODE;
int debug = 1;
int term_signal = _FAIL_SIGNAL;
char *cmdname = NULL;
int async = 0;

int timeout = 0;

int mk_child(int argc, char *argv[])
{
	pid_t childpid;
	int childstatus;
	int childcode = 0;
	struct sigaction act;
	int ret;

	childpid = fork();
	switch (childpid) {
	case -1: /* fail*/
		error("fork() failed: %s\n", strerror(errno));
		break;
	case 0: /* fork() succeeded*/
		/* kiddie*/

		argv[argc] = NULL; /* NULL terminated arg list*/
		
		/* set into own new process group*/
		setpgid(0,0);

		execvp(argv[0], argv);

		/* exec failed*/
		error("exec(%s) failed: %s\n", argv[0], strerror(errno));
		/* never returns*/
		break;
	default: /*parent*/
		/*   
		 *
		 * Set the signalhandlers for parent to ignore signals
		 * 
		 */
		act.sa_handler = SIG_IGN;
		sigemptyset(&act.sa_mask);
		act.sa_flags = 0;

		sigaction(SIGTERM, &act, NULL);
		sigaction(SIGHUP,  &act, NULL);
		sigaction(SIGINT,  &act, NULL);
		sigaction(SIGPIPE, &act, NULL);
		sigaction(SIGQUIT, &act, NULL);
		sigaction(SIGUSR1, &act, NULL);
		sigaction(SIGUSR2, &act, NULL);

		if (sleep(timeout) == 0) {
			/* kill it*/
			kill(-childpid, term_signal);
			childcode = FAIL_EXITCODE;
		} 
		do {
			ret = waitpid(childpid, &childstatus, 0);
		} while (ret == -1 && errno == EINTR);

		if (WIFEXITED(childstatus)) {
			childcode = WEXITSTATUS(childstatus);
		} else {
			childcode = FAIL_EXITCODE;
		}

		break;
	};
	return (childcode);
}

/*
 *
 * Signal Handler
 *
 */

void nop(int i) {
	return;
}

void usage() {
	fprintf(stderr, "usage: %s [-a] [-e exitcode] [-k signame] " 
		"-t secs command [args]\n", cmdname);
	exit (FAIL_EXITCODE);
}

void longusage() {
	fprintf(stderr, 
"usage: %s [-a] [-e exitcode] [-k signame] -t secs command [args]\n"
"       %s [-l|-h|-?]\n"
"Options:\n"
"  -a           Async mode. Starts %s in the background\n"
"  -e exitcode  Changes the exitcode returned by %s on fail\n"
"  -k signame   Specifies the signal witch will be sent to the process group\n"
"               if a timeout occures\n"
"  -t secs      Specifies the timeout in seconds\n"
"  -l           Print list of available signals on this platform and exit.\n"
"\n", cmdname, cmdname, cmdname, cmdname);
	fprintf(stderr, 
"Version:\n"
"  V%s\n" 
"  Copyright (c) 2001,2003 by Markus Winand <mws@fatalmind.com>\n"
"  This program is distributed in the hope that it will be useful,\n"
"  but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
"  GNU General Public License for more details.\n\n", VERSION);

	exit (FAIL_EXITCODE) ;
}


/*
 *
 * Command line parsing
 *
 */

char **cmd_line(int argc, char *argv[]) {
	char **childargv = NULL;
	int childargc = argc - 1;
	int hlp;
	char dummy;
	int ch;

#ifdef HAVE_SETENV
	setenv("POSIXLY_CORRECT", "1", 0);
#endif

	while ((ch = getopt(argc, argv, "ae:k:t:lh?")) != -1) {
		switch(ch) {
		case 'a':
			async = 1;
			break;
		case 'e': /* exitcode */
			if (sscanf(optarg, "%d%c", &hlp, &dummy) != 1) {
				error("illegal option \"-e %s\"\n", optarg);
			}		
			FAIL_EXITCODE = hlp;
			break;
			
		case 'k': /* signame */
			if (sscanf(optarg,"%d%c", &hlp, &dummy) == 1) {
				term_signal = hlp;
			} else {
				/* check for signal name*/
				hlp = signumber(optarg);
				if (hlp == 0) {
					error("unknown signal \"%s\"\n",optarg);
				}
				term_signal = hlp;
			}
			break;
		case 't':
			if (sscanf(optarg, "%d%c", &hlp, &dummy) != 1) {
				error("illegal option \"-t %s\"\n", optarg);
			}
			if (hlp <= 0) {
				error("illegal option \"-t %s\"\n", optarg);
				exit (FAIL_EXITCODE);
			}
			timeout = hlp;
			break;
		case 'l':
			siglist();
			break;
		case '?':
		case 'h':
			longusage();
			break;
		}
	}

	childargc = argc - optind;
	childargv = argv + optind;


	if (timeout == 0 || childargc < 1) {
		usage();
	}

	childargv[childargc] = NULL;

	return childargv;
}	
/*
 *
 * Main
 *
 */

int main(int argc, char *argv[]) {
	int childargc = argc - 1;
	char **childargv = NULL;
	int child_return = 99;
	struct sigaction act;
	pid_t async_fork = 0;

	cmdname = argv[0];

	if (argc < 2) {
		usage();
	}
	
	/*
	 *
	 * Command line parsing
	 *
	 */

	childargv = cmd_line(argc, argv);

	/*
	 *
	 * fork() for async mode
	 *
	 */

	if (async) {
		async_fork = fork();
		if (async_fork == -1) {
			/* fail*/
			error("fork() failed: %s\n", strerror(errno));
		}
		if (async_fork > 0) {
			/* daddy dies*/
			exit (0);
		}
	}

	/*
	 *
	 * Set a nop signalhandler for SIGCHLD, needed to interrupt the sleep
	 *
	 */
	
	act.sa_handler = nop;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &act, NULL) != 0) {
		error("sigaction() failed: %s\n", strerror(errno));
	}

	/*
	 *
	 * do the work
	 *
	 */

	child_return = mk_child(childargc, childargv);

	exit (child_return);
}


